﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using HeuristicLab.Services.Optimization.ControllerService.Interfaces;
using HeuristicLab.Optimization;
using HeuristicLab.Algorithms.GeneticAlgorithm;
using HeuristicLab.Problems.TravelingSalesman;
using HeuristicLab;
using System.Reflection;
using HeuristicLab.Services.Optimization.ControllerService.Model;
using HeuristicLab.Core;
using System.Collections;
using HeuristicLab.Clients.Hive;
using System.Threading;

namespace HeuristicLab.Services.Optimization.ControllerService {
  public class HiveScenarioManager : IScenarioManager {
    public object MapControllerDataType(object model, Parameter param) {
      switch (param.Type) {
        case ParameterType.Boolean:
          return (param.Value as Model.BoolValue).Value;
        case ParameterType.Integer:
          return Convert.ToInt32((param.Value as Model.DecimalValue).Value);          
        case ParameterType.Percent:        
        case ParameterType.Decimal:
          return (param.Value as Model.DecimalValue).Value;
        case ParameterType.DecimalMatrix:
          return Utility.ToMultiD(((param.Value as Model.DecimalMatrix).Value));
        case ParameterType.DecimalVector:
          return (param.Value as Model.DecimalVector).Value;
        case ParameterType.Type:
          //TODO handle type correctly!
          var typeValue = param.Value as Model.TypeValue;
          
          IProblem problem = model as IProblem;
          IAlgorithm algoModel = model as IAlgorithm;
          if (problem == null && algoModel != null) {
            problem = algoModel.Problem;
          }

          if (problem != null) {
            var toselect = problem.Operators.OfType<IOperator>().FirstOrDefault(x => x.Name == typeValue.Value);
            if (toselect != null) {
              return toselect;
            }
          }
          
          string parameterPropertyName = typeValue.Name.EndsWith("Parameter") ? typeValue.Name : typeValue.Name + "Parameter";
          var parameterProperty = model.GetType().GetProperty(parameterPropertyName);
          if (parameterProperty != null) {
            var parameterPropertyValue = parameterProperty.GetGetMethod().Invoke(model, null);
            var vvProp = parameterPropertyValue.GetType().GetProperty("ValidValues");
            if (vvProp != null) {
              //http://stackoverflow.com/questions/245607/how-is-generic-covariance-contra-variance-implemented-in-c-sharp-4-0 because of covariance declared in IEnumerable
              var options = vvProp.GetGetMethod().Invoke(parameterPropertyValue, null) as IEnumerable<IItem>;
              if (options != null) {
                var result = options.FirstOrDefault(x => x.ItemName == typeValue.Value);
                if (result != null)
                  return result;
              }
            }
          }

          Type type = null;
          foreach (var asm in AppDomain.CurrentDomain.GetAssemblies()) {
            type = asm.GetType(typeValue.Value);
            if (type != null)
              break;
          }
          
          return Activator.CreateInstance(type);
      }
      return null;
    }

    public object CreateHLObjectForType(object obj, Type type, Parameter prop) {
      var value = MapControllerDataType(obj, prop);
      if (prop.Type == ParameterType.Type)
        return value;

      if (type == typeof(HeuristicLab.Encodings.PermutationEncoding.Permutation)) {
        double[] input = value as double[];
        int[] arr = new int[input.Length];
        for (int i = 0; i < input.Length; i++) {
          arr[i] = Convert.ToInt32(input[i]);
        }
        var valueObj = new HeuristicLab.Encodings.PermutationEncoding.Permutation(Encodings.PermutationEncoding.PermutationTypes.RelativeUndirected, arr);
        return valueObj;
      }
      var propertyGetter = obj.GetType().GetProperty(prop.Value.Name).GetGetMethod();        
      return Activator.CreateInstance(propertyGetter.ReturnType, value);
    }

    private void MapProperty(object model, Parameter prop) {
      var propertyGetter = model.GetType().GetProperty(prop.Value.Name).GetGetMethod();
      var propertySetter = model.GetType().GetProperty(prop.Value.Name).GetSetMethod();
      var value = propertyGetter.Invoke(model, null);
      var valueProperty = value != null ? value.GetType().GetProperty("Value") : null;
      if (propertySetter != null) {
        var instance = CreateHLObjectForType(model, propertyGetter.ReturnType, prop);
        propertySetter.Invoke(model, new object[] { instance });
      }
      else if (valueProperty != null) {        
        valueProperty.GetSetMethod().Invoke(value, new object[] { MapControllerDataType(model, prop) });
      }
      else {
        Console.WriteLine("Failed to handle property!");
      }
    }

    public void DispatchScenario(Model.OptimizationScenario scenario) {
      Experiment experiment = new Experiment();
      var problem = new TravelingSalesmanProblem();
      foreach (var prop in scenario.InputParameters.Items) {
        MapProperty(problem, prop);
      }

      var algo = new GeneticAlgorithm();
      algo.Problem = problem;      
      foreach (var prop in scenario.AlgorithmParameters.Items) {
        MapProperty(algo, prop);
      }

      experiment.Optimizers.Add(algo);
      SendExperimentToHive(experiment);
    }

    private void SendExperimentToHive(Experiment exp) {
      var job = new RefreshableJob();
      job.IsAllowedPrivileged = true;      
      job.Job.Name = "Web Scheduled Traveling Salesman Job";
      job.Job.ResourceNames = "TESTGROUP";
      job.RefreshAutomatically = false;
      job.HiveTasks.Add(new OptimizerHiveTask(exp));
      HiveServiceLocator.Instance.Username = "fschoeppl";
      HiveServiceLocator.Instance.Password = "fschoeppl";
      HiveServiceLocator.Instance.EndpointConfigurationName = "WSHttpBinding_IHiveService";
      
      HiveClient.StartJob((ex) => { 
        Console.WriteLine(ex.StackTrace); 
      }, job, new CancellationToken());

      job.StopResultPolling();
    }
  }
}
